#!/usr/bin/env python3

import os, pytest, tempfile, datetime

from cozify import conftest

from cozify import cloud, config, hub
from cozify.test import debug

## basic cloud.authenticate() tests

@pytest.mark.live
def test_auth_cloud():
    assert cloud.authenticate()

@pytest.mark.live
def test_auth_hub():
    assert cloud.authenticate(trustHub=False)

class tmp_cloud():
    """Creates a temporary cloud state with test data.
    """
    def __init__(self):
        self.configfile, self.configpath = tempfile.mkstemp()
        self.section = 'Cloud'
        self.email = 'example@example.com'
        self.token = 'eyJkb20iOiJ1ayIsImFsZyI6IkhTNTEyIiwidHlwIjoiSldUIn0.eyJyb2xlIjo4LCJpYXQiOjE1MTI5ODg5NjksImV4cCI6MTUxNTQwODc2OSwidXNlcl9pZCI6ImRlYWRiZWVmLWFhYWEtYmJiYi1jY2NjLWRkZGRkZGRkZGRkZCIsImtpZCI6ImRlYWRiZWVmLWRkZGQtY2NjYy1iYmJiLWFhYWFhYWFhYWFhYSIsImlzcyI6IkNsb3VkIn0.QVKKYyfTJPks_BXeKs23uvslkcGGQnBTKodA-UGjgHg' # valid but useless jwt token.
        self.expiry = datetime.timedelta(days=1)
        self.now = datetime.datetime.now()
        self.iso_now = self.now.isoformat().split(".")[0]
        self.yesterday = self.now - datetime.timedelta(days=1)
        self.iso_yesterday = self.yesterday.isoformat().split(".")[0]
    def __enter__(self):
        config.setStatePath(self.configpath)
        cloud._setAttr('email', self.email)
        cloud._setAttr('remotetoken', self.token)
        cloud._setAttr('last_refresh', self.iso_yesterday)
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        os.remove(self.configpath)

        if exc_type is not None:
            debug.logger.error("%s, %s, %s" % (exc_type, exc_value, traceback))
            return False

@pytest.fixture
def tmpcloud(scope='module'):
    with tmp_cloud() as cloud:
        yield cloud

@pytest.fixture
def livecloud(scope='module'):
    config.setStatePath() # reset to default
    return cloud

@pytest.fixture
def id(scope='module'):
    return 'deadbeef-aaaa-bbbb-cccc-dddddddddddd'


## cloud.refresh() logic tests

def test_cloud_refresh_cold(tmpcloud):
    config.state.remove_option('Cloud', 'last_refresh')
    config.dump_state()
    assert cloud._need_refresh(force=False, expiry=tmpcloud.expiry)

def test_cloud_refresh_force(tmpcloud):
    config.dump_state()
    assert cloud._need_refresh(force=True, expiry=datetime.timedelta(days=365))

def test_cloud_refresh_expiry_over(tmpcloud):
    config.dump_state()
    assert cloud._need_refresh(force=False, expiry=datetime.timedelta(hours=1))

def test_cloud_refresh_expiry_not_over(tmpcloud):
    config.dump_state()
    assert not cloud._need_refresh(force=False, expiry=datetime.timedelta(days=2))

## integration tests for remote

@pytest.mark.live
def test_cloud_remote_match(livecloud):
    config.dump_state()
    local_tz = hub.tz()
    remote_tz = hub.tz(remote=True)

    assert local_tz == remote_tz



#!/usr/bin/env python3
import pytest

from cozify import conftest
from cozify import cloud, hub, hub_api, config

tmphub = lambda:0
tmpcloud = lambda:0

@pytest.fixture
def default_hub(scope='module'):
    config.setStatePath() # reset to default config
    config.dump_state()
    tmphub.hub_id = hub.getDefaultHub()
    tmphub.name = hub.name(tmphub.hub_id)
    tmphub.host = hub.host(tmphub.hub_id)
    tmphub.token = hub.token(tmphub.hub_id)
    tmphub.remote = hub.remote
    return tmphub

@pytest.fixture
def live_cloud(scope='module'):
    tmpcloud.token = cloud.token()
    return tmpcloud


@pytest.mark.live
def test_hub(live_cloud, default_hub):
    assert hub_api.hub(
            host = default_hub.host,
            remote = default_hub.remote,
            remote_token = live_cloud.token,
            hub_token = default_hub.token
            )



import pytest
def pytest_addoption(parser):
    parser.addoption("--live", action="store_true",
                     default=False, help="run tests requiring a functional auth and a real hub.")

def pytest_collection_modifyitems(config, items):
    if config.getoption("--live"):
        return
    skip_live = pytest.mark.skip(reason="need --live option to run")
    for item in items:
        if "live" in item.keywords:
            item.add_marker(skip_live)



import time

from . import config

# expects Cozify devices type json data
def getMultisensorData(data):
    out = []
    for device in data:
        state=data[device]['state']
        devtype = state['type']

        if devtype == 'STATE_MULTI_SENSOR':
            name=data[device]['name']
            if 'lastSeen' in state:
                timestamp=state['lastSeen']
            else:
                # if no time of measurement is known we must make a reasonable assumption
                # Stored here in milliseconds to match accuracy of what the hub will give you
                timestamp = time.time() * 1000
            if 'temperature' in state:
                temperature=state['temperature']
            else:
                temperature=None
            if 'humidity' in state:
                humidity=state['humidity']
            else:
                humidity=None

            out.append({
                    'name': name,
                    'time': timestamp,
                    'temperature': temperature,
                    'humidity': humidity
                    })
    return out



"""Module for handling highlevel Cozify Hub operations.

Attributes:
    remote(bool): Selector to treat a hub as being outside the LAN, i.e. calls will be routed via the Cozify Cloud remote call system. Defaults to False.
    autoremote(bool): Selector to autodetect hub LAN presence and flip to remote mode if needed. Defaults to True.
    capability(capability): Enum of known device capabilities. Alphabetically sorted, numeric value not guaranteed to stay constant between versions if new capabilities are added.

"""

import requests, logging
from . import config as c
from . import cloud
from . import hub_api
from enum import Enum


from .Error import APIError

remote = False
autoremote = True

capability = Enum('capability', 'BASS BRIGHTNESS COLOR_HS COLOR_LOOP COLOR_TEMP CONTACT DEVICE HUMIDITY LOUDNESS MUTE NEXT ON_OFF PAUSE PLAY PREVIOUS SEEK STOP TEMPERATURE TRANSITION TREBLE USER_PRESENCE VOLUME')

def getDevices(**kwargs):
    """Deprecated, will be removed in v0.3. Get up to date full devices data set as a dict.

    Args:
        **hub_name(str): optional name of hub to query. Will get converted to hubId for use.
        **hub_id(str): optional id of hub to query. A specified hub_id takes presedence over a hub_name or default Hub. Providing incorrect hub_id's will create cruft in your state but it won't hurt anything beyond failing the current operation.
        **remote(bool): Remote or local query.
        **hubId(str): Deprecated. Compatibility keyword for hub_id, to be removed in v0.3
        **hubName(str): Deprecated. Compatibility keyword for hub_name, to be removed in v0.3

    Returns:
        dict: full live device state as returned by the API

    """
    hub_id = _get_id(**kwargs)
    hub_token = token(hub_id)
    cloud_token = cloud.token()
    hostname = host(hub_id)

    if 'remote' not in kwargs:
        kwargs['remote'] = remote

    return devices(capability=None, **kwargs)

def devices(*, capability=None, **kwargs):
    """Get up to date full devices data set as a dict. Optionally can be filtered to only include certain devices.

    Args:
        capability(cozify.hub.capability): Capability to filter by, for example: cozify.hub.capability.TEMPERATURE. Defaults to no filtering.
        **hub_name(str): optional name of hub to query. Will get converted to hubId for use.
        **hub_id(str): optional id of hub to query. A specified hub_id takes presedence over a hub_name or default Hub. Providing incorrect hub_id's will create cruft in your state but it won't hurt anything beyond failing the current operation.
        **remote(bool): Remote or local query.
        **hubId(str): Deprecated. Compatibility keyword for hub_id, to be removed in v0.3
        **hubName(str): Deprecated. Compatibility keyword for hub_name, to be removed in v0.3

    Returns:
        dict: full live device state as returned by the API

    """
    hub_id = _get_id(**kwargs)
    hub_token = token(hub_id)
    cloud_token = cloud.token()
    hostname = host(hub_id)

    devs = hub_api.devices(host=hostname, hub_token=hub_token, remote=remote, cloud_token=cloud_token)
    if capability:
        return { key : value for key, value in devs.items() if capability.name in value['capabilities']['values'] }
    else:
        return devs

def _get_id(**kwargs):
    """Get a hub_id from various sources, meant so that you can just throw kwargs at it and get a valid id.
    If no data is available to determine which hub was meant, will default to the default hub. If even that fails, will raise an AttributeError.

    Args:
        **hub_id(str): Will be returned as-is if defined.
        **hub_name(str): Name of hub.
        hubName(str): Deprecated. Compatibility keyword for hub_name, to be removed in v0.3
        hubId(str): Deprecated. Compatibility keyword for hub_id, to be removed in v0.3
    """
    if 'hub_id' in kwargs or 'hubId' in kwargs:
        logging.debug("Redundant hub._get_id call, resolving hub_id to itself.")
        if 'hub_id' in kwargs:
            return kwargs['hub_id']
        return kwargs['hubId']
    if 'hub_name' in kwargs or 'hubName' in kwargs:
        if 'hub_name' in kwargs:
            return getHubId(kwargs['hub_name'])
        return getHubId(kwargs['hubName'])
    return getDefaultHub()

def getDefaultHub():
    """Return id of default Hub.

    If default hub isn't known, run authentication to make it known.
    """

    if 'default' not in c.state['Hubs']:
        logging.critical('no hub name given and no default known, you should run cozify.authenticate()')
        raise AttributeError
    else:
        return c.state['Hubs']['default']

def getHubId(hub_name):
    """Get hub id by it's name.

    Args:
        hub_name(str): Name of hub to query. The name is given when registering a hub to an account.

    Returns:
        str: Hub id or None if the hub wasn't found.
    """

    for section in c.state.sections():
        if section.startswith("Hubs."):
            logging.debug('Found hub {0}'.format(section))
            if c.state[section]['hubname'] == hub_name:
                return section[5:] # cut out "Hubs."
    return None

def _getAttr(hub_id, attr):
    """Get hub state attributes by attr name.

    Args:
        hub_id(str): Id of hub to query. The id is a string of hexadecimal sections used internally to represent a hub.
        attr(str): Name of hub attribute to retrieve
    Returns:
        str: Value of attribute or exception on failure.
    """
    section = 'Hubs.' + hub_id
    if section in c.state and attr in c.state[section]:
        return c.state[section][attr]
    else:
        logging.warning('Hub id "{0}" not found in state or attribute {1} not set for hub.'.format(hub_id, attr))
        raise AttributeError

def _setAttr(hub_id, attr, value, commit=True):
    """Set hub state attributes by hub_id and attr name

    Args:
        hub_id(str): Id of hub to store for. The id is a string of hexadecimal sections used internally to represent a hub.
        attr(str): Name of cloud state attribute to overwrite. Attribute will be created if it doesn't exist.
        value(str): Value to store
        commit(bool): True to commit state after set. Defaults to True.
    """
    section = 'Hubs.' + hub_id
    if section in c.state:
        if attr not in c.state[section]:
            logging.info("Attribute {0} was not already in {1} state, new attribute created.".format(attr, section))
        c.state[section][attr] = value
        if commit:
            c.stateWrite()
    else:
        logging.warning('Section {0} not found in state.'.format(section))
        raise AttributeError


def name(hub_id):
    """Get hub name by it's id.

    Args:
        hub_id(str): Id of hub to query. The id is a string of hexadecimal sections used internally to represent a hub.

    Returns:
        str: Hub name or None if the hub wasn't found.
    """
    return _getAttr(hub_id, 'hubname')

def host(hub_id):
    """Get hostname of matching hub_id

    Args:
        hub_id(str): Id of hub to query. The id is a string of hexadecimal sections used internally to represent a hub.

    Returns:
        str: ip address of matching hub. Be aware that this may be empty if the hub is only known remotely and will still give you an ip address even if the hub is currently remote.
    """
    return _getAttr(hub_id, 'host')

def token(hub_id, new_token=None):
    """Get hub_token of matching hub_id or set a new value for it.

    Args:
        hub_id(str): Id of hub to query. The id is a string of hexadecimal sections used internally to represent a hub.

    Returns:
        str: Hub authentication token.
    """
    if new_token:
        _setAttr(hub_id, 'hubtoken', new_token)
    return _getAttr(hub_id, 'hubtoken')

def ping(hub_id=None, hub_name=None, **kwargs):
    """Perform a cheap API call to trigger any potential APIError and return boolean for success/failure. For optional kwargs see cozify.hub_api.get()

    Args:
        hub_id(str): Hub to ping or default if None. Defaults to None.
        hub_name(str): Hub to ping or default if None. Defaults to None.

    Returns:
        bool: True for a valid and working hub authentication state.
    """

    if hub_name and not hub_id:
        hub_id = getHubId(hub_name)

    if not hub_id and not hub_name:
        hub_id = getDefaultHub()
    try:
        config_name = 'Hubs.' + hub_id
        hub_token = _getAttr(hub_id, 'hubtoken')
        hub_host = _getAttr(hub_id, 'host')
        cloud_token = c.state['Cloud']['remotetoken']

        # if we don't have a stored host then we assume the hub is remote
        global remote
        if not remote and autoremote and not hub_host:
            remote = True
            logging.debug('Ping determined hub is remote and flipped state to remote.')

        timezone = tz(hub_id)
        logging.debug('Ping performed with tz call, response: {0}'.format(timezone))
    except APIError as e:
        if e.status_code == 401:
            logging.debug(e)
            return False
        else:
            raise
    else:
        return True


def tz(hub_id=None, **kwargs):
    """Get timezone of given hub or default hub if no id is specified. For kwargs see cozify.hub_api.get()

    Args:
    hub_id(str): Hub to query, by default the default hub is used.

    Returns:
        str: Timezone of the hub, for example: 'Europe/Helsinki'
    """

    if not hub_id:
        hub_id = getDefaultHub()

    ip = host(hub_id)
    hub_token = token(hub_id)
    cloud_token = cloud.token()

    # if remote state not already set in the parameters, include it
    if remote not in kwargs:
        kwargs['remote'] = remote

    return hub_api.tz(host=ip, hub_token=hub_token, cloud_token=cloud_token, **kwargs)



from distutils.core import setup
from setuptools.command.test import test as TestCommand

import sys
import cozify

with open('README.rst') as file:
        long_description = file.read()

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)



setup(name='cozify',
        version = cozify.__version__,
        author = 'artanicus',
        author_email = 'python-cozify@nocturnal.fi',
        url = 'https://github.com/Artanicus/python-cozify',
        description = 'Unofficial Python bindings and helpers for the unpublished Cozify API.',
        long_description = long_description,
        license = 'MIT',
        packages = ['cozify'],
	tests_require=['pytest'],
	install_requires=['requests'],
        cmdclass={'test': PyTest},
        classifiers = [
            "Development Status :: 3 - Alpha",
            "Topic :: Utilities",
            "License :: OSI Approved :: MIT License",
            ]
        )



#!/usr/bin/env python3
import pytest

from cozify import conftest

from cozify import hub, hub_api, config, multisensor
from cozify.test import debug

class tmp_hub():
    """Creates a temporary hub section (with test data) in the current live state.
    """
    def __init__(self):
        self.id = 'deadbeef-aaaa-bbbb-cccc-dddddddddddd'
        self.name = 'HubbyMcHubFace'
        self.host = '127.0.0.1'
        self.section = 'Hubs.{0}'.format(self.id)
    def __enter__(self):
        config.setStatePath() # reset to default
        config.state.add_section(self.section)
        config.state[self.section]['hubname'] = self.name
        config.state[self.section]['host'] = self.host
        config.state['Hubs']['default'] = self.id
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            debug.logger.error("%s, %s, %s" % (exc_type, exc_value, traceback))
            return False
        config.state.remove_section(self.section)

@pytest.fixture
def tmphub(scope='module'):
    with tmp_hub() as hub:
        yield hub

@pytest.fixture
def id(scope='module'):
    return 'deadbeef-aaaa-bbbb-cccc-dddddddddddd'

@pytest.fixture
def livehub(scope='module'):
    config.setStatePath() # default config assumed to be live
    config.dump_state() # dump state so it's visible in failed test output
    assert hub.ping()
    return hub

@pytest.mark.live
def test_tz(livehub):
    assert hub.tz()

    # hand craft data needed for low-level api call hub_api.tz
    hubSection = 'Hubs.' + config.state['Hubs']['default']
    print(hub_api.tz(
        host=config.state[hubSection]['host'],
        hub_token=config.state[hubSection]['hubtoken'],
        remote=hub.remote,
        cloud_token=config.state['Cloud']['remotetoken']
        ))

def test_hub_id_to_name(tmphub):
    assert hub.name(tmphub.id) == tmphub.name

def test_hub_name_to_id(tmphub):
    assert hub.getHubId(tmphub.name) == tmphub.id

@pytest.mark.live
def test_multisensor(livehub):
    data = hub.getDevices()
    print(multisensor.getMultisensorData(data))

def test_hub_get_id(tmphub):
    assert hub._get_id(hub_id=tmphub.id) == tmphub.id
    assert hub._get_id(hub_name=tmphub.name) == tmphub.id
    assert hub._get_id(hub_name=tmphub.name, hub_id=tmphub.id) == tmphub.id
    assert hub._get_id(hubName=tmphub.name) == tmphub.id
    assert hub._get_id(hubId=tmphub.id) == tmphub.id
    assert hub._get_id() == tmphub.id
    assert not hub._get_id(hub_id='foo') == tmphub.id




"""Module for all Cozify Hub API 1:1 calls

Attributes:
    apiPath(str): Hub API endpoint path including version. Things may suddenly stop working if a software update increases the API version on the Hub. Incrementing this value until things work will get you by until a new version is published.
"""

import requests, json

from cozify import cloud_api

from .Error import APIError

apiPath = '/cc/1.7'

def _getBase(host, port=8893, api=apiPath):
    return 'http://%s:%s%s' % (host, port, api)

def _headers(hub_token):
    return { 'Authorization': hub_token }

def get(call, hub_token_header=True, base=apiPath, **kwargs):
    """GET method for calling hub API.

    Args:
        call(str): API path to call after apiPath, needs to include leading /.
        hub_token_header(bool): Set to False to omit hub_token usage in call headers.
        base(str): Base path to call from API instead of global apiPath. Defaults to apiPath.
        **host(str): ip address or hostname of hub.
        **hub_token(str): Hub authentication token.
        **remote(bool): If call is to be local or remote (bounced via cloud).
        **cloud_token(str): Cloud authentication token. Only needed if remote = True.
    """
    response = None
    headers = None
    if kwargs['remote'] and kwargs['cloud_token']:
        response = cloud_api.remote(apicall=base + call, **kwargs)
    else:
        if hub_token_header:
            headers = _headers(kwargs['hub_token'])
        response = requests.get(_getBase(host=kwargs['host'], api=base) + call, headers=headers)

    if response.status_code == 200:
        return response.json()
    elif response.status_code == 410:
        raise APIError(response.status_code, 'API version outdated. Update python-cozify. %s - %s - %s' % (response.reason, response.url, response.text))
    else:
        raise APIError(response.status_code, '%s - %s - %s' % (response.reason, response.url, response.text))

def put(call, payload, hub_token_header=True, base=apiPath, **kwargs):
    """PUT method for calling hub API.

    Args:
        call(str): API path to call after apiPath, needs to include leading /.
        payload(str): json string to push out as the payload.
        hub_token_header(bool): Set to False to omit hub_token usage in call headers.
        base(str): Base path to call from API instead of global apiPath. Defaults to apiPath.
        **host(str): ip address or hostname of hub.
        **hub_token(str): Hub authentication token.
        **remote(bool): If call is to be local or remote (bounced via cloud).
        **cloud_token(str): Cloud authentication token. Only needed if remote = True.
    """
    response = None
    headers = None
    if kwargs['remote'] and kwargs['cloud_token']:
        response = cloud_api.remote(apicall=base + call, put=True, payload=payload, **kwargs)
    else:
        if hub_token_header:
            headers = _headers(kwargs['hub_token'])
        response = requests.put(_getBase(host=kwargs['host'], api=base) + call, headers=headers, data=payload)

    if response.status_code == 200:
        return response.json()
    elif response.status_code == 410:
        raise APIError(response.status_code, 'API version outdated. Update python-cozify. %s - %s - %s' % (response.reason, response.url, response.text))
    else:
        raise APIError(response.status_code, '%s - %s - %s' % (response.reason, response.url, response.text))

def hub(**kwargs):
    """1:1 implementation of /hub API call. For kwargs see cozify.hub_api.get()

    Returns:
        dict: Hub state dict.
    """
    return get('hub', base='/', hub_token_header=False, **kwargs)

def tz(**kwargs):
    """1:1 implementation of /hub/tz API call. For kwargs see cozify.hub_api.get()

    Returns:
        str: Timezone of the hub, for example: 'Europe/Helsinki'
    """
    return get('/hub/tz', **kwargs)

def devices(**kwargs):
    """1:1 implementation of /devices API call. For kwargs see cozify.hub_api.get()

    Returns:
        json: Full live device state as returned by the API
    """
    return get('/devices', **kwargs)

def devices_command(command, **kwargs):
    """1:1 implementation of /devices/command. For kwargs see cozify.hub_api.put()

    Args:
        command(str): json string of type DeviceData containing the changes wanted

    Returns:
        str: What ever the API replied or an APIException on failure.
    """
    return put('/devices/command', command, **kwargs)
